/*	hpsum.c
 * ------------------------------------------
 *	a simple program to gain benchmarking information
 *	for executables in a parallel environment
 *
 *  written by Sean Finney <finney@cs.swarthmore.edu>
 *
 *  Copyright (C) 2001 The Regents of the University of California.  
 *  All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without
 *	modification, are permitted provided that: (1) source code distributions
 *	retain the above copyright notice and this paragraph in its entirety, (2)
 *	distributions including binary code include the above copyright notice and
 *	this paragraph in its entirety in the documentation or other materials
 *	provided with the distribution, and (3) all advertising materials mentioning
 *	features or use of this software display the following acknowledgement:
 *	``This product includes software developed by the University of California,
 *	Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 *	the University nor the names of its contributors may be used to endorse
 *	or promote products derived from this software without specific prior
 *	written permission.
 *	THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 *	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *	
 */
 // ----------------------------------------------------------------- 

#include <stdio.h>
#include <stdlib.h>		// system
#include <pthread.h>	// pthreads
#include <sys/stat.h>	// mkdir
#include <fcntl.h>		// mkdir
#include <errno.h>		// errno for mkdir
#include <unistd.h>		// sleep, mkdir, gettimeofday
#include <sys/types.h>	// time, mkdir
#include <time.h>		// time
#include <string.h>		// strtok
#include <pmapi.h>		// duh
#include <sys/time.h>	// gettimeofday

// ----------------------------------------------------------------- 
	
#define UNSET_VALUE -1
#define DEFAULT_INTERVAL 1000
#define MAX_HOSTNAMESIZE 64
#define MAX_INT_STRLEN 10		// 32 bits == ~4.0 * 10^9, so 10 chars
#define SUBDIR_PREFIX ".hpsum-"
#define SUBDIR_PREFIX_LEN 8
#define SUFFIX ".hpsum"
#define SUFFIX_LEN 6	
#define SUMMARY_SCRIPT "hpsum-summary.pl"

#define MIN_CMDLINE_ARGS 2

// These are for bail, sort of frivolous but makes it a little nicer
//	for whether or not to call usage
#define INTERNAL_ERROR -2
#define USER_ERROR -1

// ----------------------------------------------------------------- 

extern int errno;		// for mkdir
extern char **environ;		// uses environ to pick unique dirnames
char* g_progname;		// argv[0]
char* g_subdir;			// unique subdir name
char* g_userprog;		// prog to be run, with args
long g_interval = UNSET_VALUE;
short g_fpux_reporter = UNSET_VALUE;
short g_data_at_t = UNSET_VALUE;
short g_bequiet = UNSET_VALUE;
char g_filename[MAX_INT_STRLEN+MAX_HOSTNAMESIZE+MAX_INT_STRLEN+SUFFIX_LEN+1];
short g_complete = 0;

// ----------------------------------------------------------------- 

void parse_cmdline(int count, char* args[]);
void* do_program(void *arg);
void bail(char *msg, int status);
void usage(void);
void touch_pid_file(pid_t pid);
void pick_master_task(pid_t pid);
void do_epilogue(void);

// ----------------------------------------------------------------- 

int main(int argc, char *argv[]){
	int i=0;
	pm_info_t my_info;
	pm_prog_t my_prog;
	pm_data_t my_data;
	long long *oldvals;
	char hostname[MAX_HOSTNAMESIZE];
	pid_t pid;
	pthread_t my_tid, doprog_tid;
	pthread_attr_t attr;
	long long time=0, timeleft=0;
	FILE *outfile;
	char *partition;
	struct timespec nanotime, nanorem;

	my_tid=pthread_self();
	parse_cmdline(argc, argv);
	partition=getenv("MP_PARTITION");
	if(!partition) bail("MP_PARTITION must be defined!", INTERNAL_ERROR); 
	g_subdir=(char*)malloc(sizeof(char)*
					(SUBDIR_PREFIX_LEN+strlen(partition)+1));
	nanotime.tv_sec=g_interval/1000;
	nanotime.tv_nsec=(g_interval%1000)*1000000;

	sprintf(g_subdir, "%s%s", SUBDIR_PREFIX, partition);
	if(mkdir(g_subdir, S_IRWXU)==-1 && errno!=EEXIST) 
		perror("mkdir: ");

	gethostname(hostname, MAX_HOSTNAMESIZE);
	strtok(hostname, ".");
	pid = getpid();
	sprintf(g_filename, "%s.%d%s", hostname, pid, SUFFIX);
	touch_pid_file(pid);
	sprintf(g_filename, "%s/%s.%d%s", g_subdir, hostname, pid, SUFFIX);
	outfile=fopen(g_filename, "w");
	if(!outfile) bail("unable to open file for writing", INTERNAL_ERROR);

	my_prog.mode.w			= 0;
	my_prog.mode.b.user		= 1;
	my_prog.mode.b.kernel	= 0;
	my_prog.mode.b.proctree	= 1;

	pm_init(PM_VERIFIED, &my_info);

	oldvals=(long long*)malloc(sizeof(long long)*(my_info.maxpmcs));
	for(i=0;i<my_info.maxpmcs; i++)	oldvals[i] = 0;
	for(i=0;i<my_info.maxpmcs; i++) my_prog.events[i]=COUNT_NOTHING;

	my_prog.events[0] = 0;		// 0,0==Processor Cycles (PM_CYC)
	my_prog.events[1] = 35;	// 1,35==fpu1 insts (PM_FPU1_CMPL)
	my_prog.events[2] = 2; // 2,2==Instructions Complete (PM_INST_CMPL)
	my_prog.events[4] = 5;  // 4,5==fpu0 insts	(PM_FPU0_CMPL)
	my_prog.events[6] = 9; 	// 6,9==Float multiply-adds (PM_EXEC_FMA)
	my_prog.events[7] = 2;	// 7,2==L2 cach miss [b] (PM_BURSTRD_L2MISS)

	//my_prog.events[0] = 19;	// 0,19==TLB misses/digs to mem (PM_TLB_MISS)
	//my_prog.events[5] = 12; // 5,12==processor cycles (PM_CYC)
	//my_prog.events[6] = 3; 	// 6,3==mispred. branch (PM_MPRED_BR_CAUSED_GC)
	//my_prog.events[2] = 12; // 2,12==L2 cache miss [a] (PM_ST_L2MISS)
	//my_prog.events[3] = 5;	// 3,5==loads completed (PM_LD_CMPL)

	pm_set_program_mygroup(&my_prog);
	pm_reset_data_mygroup();
	pm_start_mygroup();

	pthread_attr_init(&attr);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	pm_reset_data_mygroup();
	pthread_create(&doprog_tid, &attr, do_program, &my_tid);
	
	while(pthread_kill(doprog_tid, 0) == 0){
		pm_get_data_mygroup(&my_data);
		timeleft=0;
		fprintf(outfile, "%lld:%s:%d::PM_FPU0_CMPL:%lld:%lld\n", 
			time, hostname, pid, my_data.accu[4], my_data.accu[4]-oldvals[4]);
		fprintf(outfile, "%lld:%s:%d::PM_FPU1_CMPL:%lld:%lld\n", 
			time, hostname, pid, my_data.accu[1], my_data.accu[1]-oldvals[1]);
		fprintf(outfile, "%lld:%s:%d::PM_INST_CMPL:%lld:%lld\n", 
			time, hostname, pid, my_data.accu[2], my_data.accu[2]-oldvals[2]);
		fprintf(outfile, "%lld:%s:%d::PM_BURSTRD_L2MISS:%lld:%lld\n", 
			time, hostname, pid, my_data.accu[7], my_data.accu[7]-oldvals[7]);
		fprintf(outfile, "%lld:%s:%d::PM_CYC:%lld:%lld\n", 
			time, hostname, pid, my_data.accu[0], my_data.accu[0]-oldvals[0]);
		// ----------XXX----------
		// this has to be the last one reported, because it's not reported
		//	in the final printing due to some kind of pmapi bug.  if it's the
		//	last one written here, it can still be counted by the script.
		fprintf(outfile, "%lld:%s:%d::PM_EXEC_FMA:%lld:%lld\n", 
			time, hostname, pid, my_data.accu[6], my_data.accu[6]-oldvals[6]);

		if(nanosleep(&nanotime, &nanorem)==-1){
			timeleft = (nanorem.tv_nsec/1000000) + (nanorem.tv_sec*1000);
		}
		time += ((nanotime.tv_sec*1000)+(nanotime.tv_nsec/1000000)-timeleft);

		for(i=0;i<my_info.maxpmcs; i++) oldvals[i]=my_data.accu[i];
	}

	pm_stop_mygroup();
	pm_get_data_mygroup(&my_data);

	fprintf(outfile, "%lld:%s:%d::PM_FPU0_CMPL:%lld:%lld\n", 
		time, hostname, pid, my_data.accu[4], my_data.accu[4]-oldvals[4]);
	fprintf(outfile, "%lld:%s:%d::PM_FPU1_CMPL:%lld:%lld\n", 
		time, hostname, pid, my_data.accu[1], my_data.accu[1]-oldvals[1]);
	fprintf(outfile, "%lld:%s:%d::PM_INST_CMPL:%lld:%lld\n", 
		time, hostname, pid, my_data.accu[2], my_data.accu[2]-oldvals[2]);
	fprintf(outfile, "%lld:%s:%d::PM_BURSTRD_L2MISS:%lld:%lld\n", 
		time, hostname, pid, my_data.accu[7], my_data.accu[7]-oldvals[7]);
	fprintf(outfile, "%lld:%s:%d::PM_CYC:%lld:%lld\n", 
		time, hostname, pid, my_data.accu[0], my_data.accu[0]-oldvals[0]);
	//fprintf(outfile, "%lld:%s:%d::PM_EXEC_FMA:%lld:%lld\n", 
	//	time, hostname, pid, oldvals[6], oldvals[6]);
	fprintf(outfile, "--FIN--\n");

	fflush(outfile);
	fflush(stdout);
	fclose(outfile);

	pick_master_task(pid);
	pm_reset_data_mygroup();
	exit(0);
}

// ----------------------------------------------------------------- 

void touch_pid_file(pid_t pid){
	int result;
	char *cmd="touch ";
	char *prefix="pid.";
	char *bar=(char*)malloc(sizeof(char)*
			(strlen(cmd)+strlen(prefix)+strlen(g_filename)+1));
	sprintf(bar, "%s%s/%s%s", cmd, g_subdir, prefix, g_filename);
	result = system(bar);
	if(result==127 || result == -1) 
		bail("unable to touch pid_file!", INTERNAL_ERROR);	
}

void pick_master_task(pid_t pid){
	FILE *foo;
	char *cmd_prefix="ls -1 ";
	char *cmd_suffix="pid.*";
	char dump[SUBDIR_PREFIX_LEN+MAX_INT_STRLEN+1];
	char *first_filename = (char*)malloc(sizeof(char)*(strlen(g_filename)+1));
	char *cmd = (char*)malloc(sizeof(char)*
			(strlen(cmd_prefix)+strlen(g_subdir)+1+strlen(cmd_suffix)+1));

	sprintf(cmd, "%s%s/%s", cmd_prefix, g_subdir, cmd_suffix);
	foo=popen(cmd, "r");

	if(!foo) bail("unable to popen() ls -1!", INTERNAL_ERROR);

	fscanf(foo, "%[^/]/pid.%s", dump, first_filename);
	pclose(foo);

	if(strcmp(strstr(g_filename, first_filename), first_filename)==0) 
		do_epilogue();
}

void do_epilogue(void){
	char *cmd_prefix="rm -rf ";
	char *cmd;
	int retval;

	retval = system(SUMMARY_SCRIPT);
	if(retval==-1 || retval==127) perror("unable to do epilogue script: ");
	else if(g_bequiet){
		cmd=(char*)malloc(sizeof(char)*(strlen(cmd_prefix)+strlen(g_subdir)+1));
		sprintf(cmd, "%s%s", cmd_prefix, g_subdir);
		retval = system(cmd);
		if(retval==-1 || retval==127) perror("unable to do cleanup: ");
	}
	else printf("done.  output can be found in: %s\n", g_subdir);
}

void parse_cmdline(int count, char *args[]){
	int pos=1;
	int interval_modifier=1;	// if they wanna do minutes/hours etc
	int user_args_start;
	int user_args_len=0;		// len of cmdline opts to the actual prog

	g_progname=args[0];

	if(count < MIN_CMDLINE_ARGS) 
		bail("wrong number of args, fool!", USER_ERROR);

	while(pos < count){
		// -im/-is/-iM/-ih for milliseconds seconds minutes or hours
		// default: 1 second (-is 1)
		if(strncmp("-i", args[pos], 2)==0){
			if(strlen(args[pos])>2){
				switch(args[pos][2]){
				case 'm': break;
				case 's': interval_modifier=1000; break;
				case 'M': interval_modifier=60000; break;
				case 'h': interval_modifier=3600000; break;
				default: bail("invalid time modifier for -i", USER_ERROR);
				}
			} 
			if(++pos > count) bail("-i requires interval period", USER_ERROR);
			g_interval=atoi(args[pos])*interval_modifier;
			if(!g_interval) bail("invalid interval for -i", USER_ERROR);
			pos++;
			continue;
		}
		//	-q/-v quiet or verbose.  quiet removes the data dir when it's
		//		done, verbose leaves it for analyzing.
		//	default: quiet
		if(strncmp("-q", args[pos], 2)==0||strncmp("-v", args[pos], 2)==0){ 
			if(args[pos][1]=='v') g_bequiet=0;
			else if(args[pos][1]=='q') g_bequiet=1; 
			else bail("error in processing -q/-v command", INTERNAL_ERROR);
			pos++;
			continue;
		}


		// anything else should go here

		// and the args[pos] matches nothing else, the rest is assumed
		//	to be the program name and its cmdline opts, so copy it over
		//	to g_userprog
		else{
			user_args_start=pos;
			while(pos<count) user_args_len+=strlen(args[pos++]); 
			pos++;
			g_userprog=(char*)malloc(sizeof(char)*(user_args_len+count-pos+1));
			g_userprog[0]='\0';
			while(user_args_start<count){
				strcat(g_userprog, args[user_args_start++]);
				strcat(g_userprog, " ");
			}
			g_userprog[strlen(g_userprog)]='\0';	// bybye trailing space
		}
	}
	if(g_interval==UNSET_VALUE) g_interval=DEFAULT_INTERVAL;
	if(g_bequiet==UNSET_VALUE) g_bequiet=1;
}

void* do_program(void *arg) {
	int retval=system(g_userprog);
	g_complete=1;
	if(retval==-1 || retval==127)
		bail("call to system() failed--possible typo?", USER_ERROR);
	pthread_exit(0);
}

void bail(char *msg, int status){
	fprintf(stderr, "%s:\t%s\n", g_progname, msg);
	if(status==USER_ERROR) usage();
	exit(status);
}

void usage(void){
fprintf(stderr, "usage:\n");
fprintf(stderr, "%s [-i[m|s|M|h] time] program_name [progam_args]\n",
		g_progname);
}
